/*---------------------------------------------------------------
           FILE: cor.c
      $Revision: 2.4 $
         $State: Exp $
  Last submitted $Date: 94/04/14 20:23:21 $ $Author: jh $

   Locked out by $Locker: jh $

  REVISION HISTORY
  $Log:	cor.c,v $
 * Revision 2.4  94/04/14  20:23:21  jh
 * Corrected inverse product correlations to be negative, not reciprocal.
 * Reciprocal of negative products doesn't make them larger than positive
 * products...
 * 
 * Revision 2.3  93/05/14  00:42:25  jh
 * Added iprodsum cor
 * 
 * Revision 2.2  93/05/13  19:37:27  jh
 * Eliminated extraneous inclusion of string.h
 * 
 * Revision 2.1  93/05/13  19:26:24  jh
 * Eliminated extraneous definition of Streq() macro
 * 
 * Revision 2.0  93/03/30  18:00:14  jh
 * Actual first release version.
 * 
 * Revision 1.1  93/03/30  13:59:09  jh
 * General code cleanup, testing, and added features.
 * First release version.
 * 
 * Revision 1.1  92/12/15  13:08:39  jh
 * Initial revision
 * 

  $Source: /u3/jh/jupobs92/anal/iraf/src/jiggle/RCS/cor.c,v $

  Copyright (c) 1992, 1993 Joseph Harrington.
  All Rights reserved.
---------------------------------------------------------------*/
#ifndef LINT
static char rcsid[] = "$Header: /u3/jh/jupobs92/anal/iraf/src/jiggle/RCS/cor.c,v 2.4 94/04/14 20:23:21 jh Exp Locker: jh $";
static char copyright[] =
    "Copyright (c) 1992, 1993 Joseph Harrington.  All Rights reserved.";
#endif

/*---------------------------------------------------------------
cor.c = functions to do correlations on pairs of images.

External object files or libraries:
memory.c

COMMON DOCUMENTATION:

Effect: All functions take Doub2 *a and *b and return a pointer
        to a Doub2 * which is the correlation image of a with b.
        This is generally used as an evaluation of the quality
        of candidate registrations of the two images in a mosaic.
        The resulting image dimensions are one less than the sum
        in each direction of the dimensions of the input images.
        Pixel values in this image are calculated as follows:

	The images are offset from one another according to the
        coordinates of the pixel whose value is being
        calculated, their overlaps are computed, corresponding
        pixels in the overlap regions are operated upon, and the
        mean of this value in the overlap becomes the pixel
        value in the output image.

	The offset is such that the lower left corner of the
        output image corresponds to the upper right corner of b
        overlapping the lower left corner of a by one pixel in
        each dimension.

	Because one usually has some idea of the correct offset
	(called the "nominal registration"), one often wants to
	eliminate other, incorrect, solutions from consideration
	in a mosaic, and because calculating the full correlation
	is compute-intensive, one can enter a nominal
	registration of b with respect to a, and a radius from
	this location within which the correlation will be
	calculated.  Set radius to DBL_MAX to get the full
	correlation.

Requires: a, b: pointers to Doub2 containing input images for
                correlation.

	  nomx, nomy: doubles which are coordinates of
                      nominal registration.  0,0 corresponds to
		      first pixel in each image registered.

	  radius: double precision distance from nominal
	          registration for which to calculate
	          correlation.  To ignore nomx, nomy, and
	          radius, and calculate all pixel values, set
	          radius to DBL_MAX (or any sufficiently large
		  number).

Modifies: nothing.

Implementation notes: Free the returned memory with free_doub2.

                      To produce images viewable with IRAF,
		      define IRAF or the values will not be legal
		      in IRAF images. 
---------------------------------------------------------------*/
/* definitions */

#define Min(a, b) ((a) < (b) ? (a) : (b))
#define Max(a, b) ((a) > (b) ? (a) : (b))

#ifdef IRAF
#define OUTOFBOUNDS 1.6e38
#else
#define OUTOFBOUNDS DBL_MAX
#endif

/* inclusions */

#include "memory.h"
#include "cor.h"

/*----------------------snscor-----------------------------------
Correlation: Squared noise-to-signal.  Subtract corresponding
        pixels, divide by their mean value, square.
---------------------------------------------------------------*/
Doub2 *
snscor(
      Doub2 *a, 
      Doub2 *b,
      double nomx,
      double nomy,
      double radius
){
    int aw = 0;
    int ah = 0;
    int bw = 0;
    int bh = 0;

    int outw = 0;
    int outh = 0;
    int outx = 0;
    int outy = 0;

    int xoff = 0;
    int yoff = 0;

    int oal = 0;
    int oar = 0;
    int oab = 0;
    int oat = 0;

    int obl = 0;
    int obb = 0;

    int ow = 0;
    int oh = 0;

    int ox = 0;
    int oy = 0;

    double av = 0.;
    double bv = 0.;
    double xdist = 0.;
    double ydist = 0.;
    double ns = 0.;
    double sum = 0.;

    Doub2 *out;

    /* notational convenience */

    aw = a->nj;
    ah = a->ni;
    bw = b->nj;
    bh = b->ni;

    outw = aw + bw - 1;
    outh = ah + bh - 1;

    /* allocate output memory */
    out = alloc_doub2(outh, outw);
    if (!out) {
	return NULL;
    }

    /* for each pixel in the output image */

    for     ( yoff = -bh + 1 , outy = 0 ; yoff < ah ; yoff++ , outy++ ){
	for ( xoff = -bw + 1 , outx = 0 ; xoff < aw ; xoff++ , outx++ ){

	    /* determine if point is within range being considered */
	    xdist = (double) xoff - nomx;
	    ydist = (double) yoff - nomy;
	    if ( xdist * xdist + ydist * ydist < radius * radius ){

		/* calculate bounds & size of overlap regions */
		/* items commented out are here for completeness */

		oal = Max(0,       xoff);
		oar = Min(aw - 1,  xoff + bw - 1);
		oab = Max(0,       yoff);
		oat = Min(ah - 1,  yoff + bh - 1);

		obl = Max(0,      -xoff);
/*	        obr = Min(bw - 1, -xoff + aw - 1); */
		obb = Max(0,      -yoff);
/*	        obt = Min(bh - 1, -yoff + ah - 1); */

		ow = oar - oal + 1;
		oh = oat - oab + 1;

		/* for each pixel in overlap region */

		for ( sum = 0 , oy = 0 ; oy < oh ; oy++ ){
		    for       ( ox = 0 ; ox < ow ; ox++ ){

			/* calculate correlation */
			av = a->data[(oab + oy)][(oal + ox)];
			bv = b->data[(obb + oy)][(obl + ox)];
			ns = (bv - av) / ((bv + av) * 0.5);
			sum += ns * ns;
		    }
		}
		out->data[outy][outx] = sum/(double)(ow * oh);
	    } else {
		out->data[outy][outx] = OUTOFBOUNDS;
	    }
	}
    }

    /* return */

    return out;
}

/*----------------------sncor------------------------------------
Correlation: Squared noise.  Subtract corresponding pixels,
             square.
---------------------------------------------------------------*/
Doub2 *
sncor(
      Doub2 *a, 
      Doub2 *b,
      double nomx,
      double nomy,
      double radius
){
    int aw = 0;
    int ah = 0;
    int bw = 0;
    int bh = 0;

    int outw = 0;
    int outh = 0;
    int outx = 0;
    int outy = 0;

    int xoff = 0;
    int yoff = 0;

    int oal = 0;
    int oar = 0;
    int oab = 0;
    int oat = 0;

    int obl = 0;
    int obb = 0;

    int ow = 0;
    int oh = 0;

    int ox = 0;
    int oy = 0;

    double xdist = 0.;
    double ydist = 0.;
    double ns = 0.;
    double sum = 0.;

    Doub2 *out;

    /* notational convenience */

    aw = a->nj;
    ah = a->ni;
    bw = b->nj;
    bh = b->ni;

    outw = aw + bw - 1;
    outh = ah + bh - 1;

    /* allocate output memory */
    out = alloc_doub2(outh, outw);
    if (!out) {
	return NULL;
    }

    /* for each pixel in the output image */

    for     ( yoff = -bh + 1 , outy = 0 ; yoff < ah ; yoff++ , outy++ ){
	for ( xoff = -bw + 1 , outx = 0 ; xoff < aw ; xoff++ , outx++ ){

	    /* determine if point is within range being considered */
	    xdist = (double) xoff - nomx;
	    ydist = (double) yoff - nomy;
	    if ( xdist * xdist + ydist * ydist < radius * radius ){

		/* calculate bounds & size of overlap regions */
		/* items commented out are here for completeness */

		oal = Max(0,       xoff);
		oar = Min(aw - 1,  xoff + bw - 1);
		oab = Max(0,       yoff);
		oat = Min(ah - 1,  yoff + bh - 1);

		obl = Max(0,      -xoff);
/*	        obr = Min(bw - 1, -xoff + aw - 1); */
		obb = Max(0,      -yoff);
/*	        obt = Min(bh - 1, -yoff + ah - 1); */

		ow = oar - oal + 1;
		oh = oat - oab + 1;

		/* for each pixel in overlap region */

		for ( sum = 0 , oy = 0 ; oy < oh ; oy++ ){
		    for       ( ox = 0 ; ox < ow ; ox++ ){

			/* calculate correlation */
			ns =   a->data[(oab + oy)][(oal + ox)]
			     - b->data[(obb + oy)][(obl + ox)];
			sum += ns * ns;
		    }
		}
		out->data[outy][outx] = sum/(double)(ow * oh);
	    } else {
		out->data[outy][outx] = OUTOFBOUNDS;
	    }
	}
    }

    /* return */

    return out;
}

/*----------------------iprodcor---------------------------------
Correlation: Inverse product.  Multiply corresponding pixels,
             sum, take negative of average.
---------------------------------------------------------------*/
Doub2 *
iprodcor(
      Doub2 *a, 
      Doub2 *b,
      double nomx,
      double nomy,
      double radius
){
    int aw = 0;
    int ah = 0;
    int bw = 0;
    int bh = 0;

    int outw = 0;
    int outh = 0;
    int outx = 0;
    int outy = 0;

    int xoff = 0;
    int yoff = 0;

    int oal = 0;
    int oar = 0;
    int oab = 0;
    int oat = 0;

    int obl = 0;
    int obb = 0;

    int ow = 0;
    int oh = 0;

    int ox = 0;
    int oy = 0;

    double xdist = 0.;
    double ydist = 0.;
    double sum = 0.;

    Doub2 *out;

    /* notational convenience */

    aw = a->nj;
    ah = a->ni;
    bw = b->nj;
    bh = b->ni;

    outw = aw + bw - 1;
    outh = ah + bh - 1;

    /* allocate output memory */
    out = alloc_doub2(outh, outw);
    if (!out) {
	return NULL;
    }

    /* for each pixel in the output image */

    for     ( yoff = -bh + 1 , outy = 0 ; yoff < ah ; yoff++ , outy++ ){
	for ( xoff = -bw + 1 , outx = 0 ; xoff < aw ; xoff++ , outx++ ){

	    /* determine if point is within range being considered */
	    xdist = (double) xoff - nomx;
	    ydist = (double) yoff - nomy;
	    if ( xdist * xdist + ydist * ydist < radius * radius ){

		/* calculate bounds & size of overlap regions */
		/* items commented out are here for completeness */

		oal = Max(0,       xoff);
		oar = Min(aw - 1,  xoff + bw - 1);
		oab = Max(0,       yoff);
		oat = Min(ah - 1,  yoff + bh - 1);

		obl = Max(0,      -xoff);
/*	        obr = Min(bw - 1, -xoff + aw - 1); */
		obb = Max(0,      -yoff);
/*	        obt = Min(bh - 1, -yoff + ah - 1); */

		ow = oar - oal + 1;
		oh = oat - oab + 1;

		/* for each pixel in overlap region */

		for ( sum = 0 , oy = 0 ; oy < oh ; oy++ ){
		    for       ( ox = 0 ; ox < ow ; ox++ ){

			/* calculate correlation */
			sum +=   a->data[(oab + oy)][(oal + ox)]
			       * b->data[(obb + oy)][(obl + ox)];
		    }
		}
		out->data[outy][outx] = -sum/(double)(ow * oh);
	    } else {
		out->data[outy][outx] = OUTOFBOUNDS;
	    }
	}
    }

    /* return */

    return out;
}

/*----------------------iprodsumcor---------------------------------
Correlation: Inverse product.  Multiply corresponding pixels,
             sum, take negative (of total, not average).
---------------------------------------------------------------*/
Doub2 *
iprodsumcor(
      Doub2 *a, 
      Doub2 *b,
      double nomx,
      double nomy,
      double radius
){
    int aw = 0;
    int ah = 0;
    int bw = 0;
    int bh = 0;

    int outw = 0;
    int outh = 0;
    int outx = 0;
    int outy = 0;

    int xoff = 0;
    int yoff = 0;

    int oal = 0;
    int oar = 0;
    int oab = 0;
    int oat = 0;

    int obl = 0;
    int obb = 0;

    int ow = 0;
    int oh = 0;

    int ox = 0;
    int oy = 0;

    double xdist = 0.;
    double ydist = 0.;
    double sum = 0.;

    Doub2 *out;

    /* notational convenience */

    aw = a->nj;
    ah = a->ni;
    bw = b->nj;
    bh = b->ni;

    outw = aw + bw - 1;
    outh = ah + bh - 1;

    /* allocate output memory */
    out = alloc_doub2(outh, outw);
    if (!out) {
	return NULL;
    }

    /* for each pixel in the output image */

    for     ( yoff = -bh + 1 , outy = 0 ; yoff < ah ; yoff++ , outy++ ){
	for ( xoff = -bw + 1 , outx = 0 ; xoff < aw ; xoff++ , outx++ ){

	    /* determine if point is within range being considered */
	    xdist = (double) xoff - nomx;
	    ydist = (double) yoff - nomy;
	    if ( xdist * xdist + ydist * ydist < radius * radius ){

		/* calculate bounds & size of overlap regions */
		/* items commented out are here for completeness */

		oal = Max(0,       xoff);
		oar = Min(aw - 1,  xoff + bw - 1);
		oab = Max(0,       yoff);
		oat = Min(ah - 1,  yoff + bh - 1);

		obl = Max(0,      -xoff);
/*	        obr = Min(bw - 1, -xoff + aw - 1); */
		obb = Max(0,      -yoff);
/*	        obt = Min(bh - 1, -yoff + ah - 1); */

		ow = oar - oal + 1;
		oh = oat - oab + 1;

		/* for each pixel in overlap region */

		for ( sum = 0 , oy = 0 ; oy < oh ; oy++ ){
		    for       ( ox = 0 ; ox < ow ; ox++ ){

			/* calculate correlation */
			sum +=   a->data[(oab + oy)][(oal + ox)]
			       * b->data[(obb + oy)][(obl + ox)];
		    }
		}
		out->data[outy][outx] = -sum;
	    } else {
		out->data[outy][outx] = OUTOFBOUNDS;
	    }
	}
    }

    /* return */

    return out;
}
